from abc import abstractmethod

from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration
from flamapy.core.operations import Operation


class Sampling(Operation):
    """Sampling is the selection of a subset (i.e., a sample) of products
    (or configurations) from within a variability model.
    """

    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def set_sample_size(self, sample_size: int) -> None:
        """Number of configurations of the sample."""

    @abstractmethod
    def set_with_replacement(self, with_replacement: bool) -> None:
        "Allow repeated configurations in the sample (default False)."

    @abstractmethod
    def set_partial_configuration(self, partial_configuration: Configuration) -> None:
        "From which the sample is built (default empty configuration)."

    @abstractmethod
    def get_sample(self) -> list[Configuration]:
        """Return a sample of configurations."""



from abc import abstractmethod

from flamapy.core.models import VariabilityModel
from flamapy.core.transformations import Transformation


class ModelToText(Transformation):

    @staticmethod
    @abstractmethod
    def get_destination_extension() -> str:
        pass

    @abstractmethod
    def __init__(self, path: str, source_model: VariabilityModel) -> None:
        pass

    @abstractmethod
    def transform(self) -> str:
        pass



from .abstract_operation import Operation
from .average_branching_factor import \
    AverageBranchingFactor  # pylint: disable=cyclic-import
from .commonality import Commonality  # pylint: disable=cyclic-import
from .core_features import CoreFeatures  # pylint: disable=cyclic-import
from .count_leafs import CountLeafs  # pylint: disable=cyclic-import
from .dead_features import DeadFeatures  # pylint: disable=cyclic-import
from .error_detection import ErrorDetection  # pylint: disable=cyclic-import
from .error_diagnosis import ErrorDiagnosis  # pylint: disable=cyclic-import
from .false_optional_features import \
    FalseOptionalFeatures  # pylint: disable=cyclic-import
from .filter import Filter  # pylint: disable=cyclic-import
from .configurations import Configurations  # pylint: disable=cyclic-import
from .configurations_number import ConfigurationsNumber  # pylint: disable=cyclic-import
from .estimated_configurations_number import \
    EstimatedConfigurationsNumber  # pylint: disable=cyclic-import
from .satisfiable import Satisfiable  # pylint: disable=cyclic-import
from .satisfiable_configuration import \
    SatisfiableConfiguration  # pylint: disable=cyclic-import
from .variability import Variability  # pylint: disable=cyclic-import
from .sampling import Sampling  # pylint: disable=cyclic-import
from .atomic_sets import AtomicSets  # pylint: disable=cyclic-import
from .metrics_operation import Metrics  # pylint: disable=cyclic-import

__all__ = ["Commonality", "DeadFeatures", "CoreFeatures", "FalseOptionalFeatures", 
           "ErrorDetection", "ErrorDiagnosis", "Operation", "Variability", "CountLeafs", 
           "AverageBranchingFactor", "Filter", "Sampling", 
           "EstimatedConfigurationsNumber", "AtomicSets", 'Metrics', 
           "Satisfiable", "SatisfiableConfiguration", 
           "Configurations", "ConfigurationsNumber"]


from abc import abstractmethod
from typing import Any

from flamapy.core.operations import Operation


class AtomicSets(Operation):
    """An atomic set is a group of features that can be treated as a unit when 
    performing certain operations.

    The intuitive idea of atomic sets is that mandatory features and their parent features 
    can be treated as a whole in certain analysis operation without altering the result. 
    This is because those features can never appear in a product separately.
    """

    @abstractmethod
    def atomic_sets(self) -> list[set[Any]]:
        """Return a list of atomic sets of the feature model."""  



from .variability_model import VariabilityModel, VariabilityElement  # pylint: disable=cyclic-import
from .ast import AST, ASTOperation  # pylint: disable=cyclic-import

__all__ = ["VariabilityModel",  "AST", "VariabilityElement", "ASTOperation"]



from abc import abstractmethod

from flamapy.core.operations import Operation


class CountLeafs(Operation):

    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_number_of_leafs(self) -> int:
        pass



from typing import Any, Iterator
from flamapy.core.models import VariabilityModel


class Configuration(VariabilityModel):

    @staticmethod
    def get_extension() -> str:
        return 'configuration'

    def __init__(self, elements: dict[Any, bool]) -> None:
        self.elements = elements
        self.is_full = False

    def set_full(self, is_full: bool) -> None:
        self.is_full = is_full

    def get_selected_elements(self) -> list[Any]:
        return [e for e in self.elements if self.elements[e]]

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Configuration):
            return self.elements == other.elements
        return False

    def __hash__(self) -> int:
        return hash(frozenset(self.elements.items()))

    def __str__(self) -> str:
        return ', '.join([str(e) for e in self.get_selected_elements()])

    def __repr__(self) -> str:
        return f"Configuration({self.elements})"

    def __iter__(self) -> Iterator[Any]:
        return iter(self.elements)



from abc import abstractmethod

from flamapy.core.operations import Operation
from flamapy.core.models.variability_model import VariabilityElement


class DeadFeatures(Operation):

    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_dead_features(self) -> list[VariabilityElement]:
        pass



from abc import abstractmethod

from flamapy.core.operations import Operation
from flamapy.core.models.variability_model import VariabilityElement


class FalseOptionalFeatures(Operation):
    """A feature is defined as `false optional` if the selection of its parent makes the feature
    itself selected as well, although it is defined as optional and not mandatory.

    A feature in a group can be also `false optional`.
    In general, false optional features should be marked as mandatory for maintainability purposes.
    """

    @abstractmethod
    def get_false_optional_features(self) -> list[VariabilityElement]:
        pass



from abc import abstractmethod, ABCMeta
import logging
import json 

from typing import Any, Optional, Collection, Type

from flamapy.core.exceptions import FlamaException
from flamapy.core.transformations.model_to_model import ModelToModel
from flamapy.core.operations import Operation
from flamapy.core.models import VariabilityModel

LOGGER = logging.getLogger('Metrics')


class Metrics(Operation, metaclass=ABCMeta):
    """This is intended to host a set of metrics calculations for a variability model.

    This abstract class will recruit all its implementations and agregate the results
    """
    filter: Optional[list[str]] = None
    result: list[dict[str, Any]] = []

    def __init__(self) -> None:
        self.model: Optional[VariabilityModel] = None

    @abstractmethod
    def calculate_metamodel_metrics(self, model: VariabilityModel) -> list[dict[str, Any]]:
        """Return a list of metrics for each metamodel."""  

    @property
    @abstractmethod
    def model_type_extension(self) -> str:
        """Return the model type extension for the specific metric."""

    def only_these_metrics(self, filter_metrics: list[str]) -> None:
        self.filter = filter_metrics 

    def execute(self, model: VariabilityModel) -> 'Metrics':
        self.model = model
        # Identifying all implementations of MetricsOperation

        for subclass in Metrics.__subclasses__(): 
            # We first have to identify the metamodels that are being used and 
            # transform this model to the correspointing metamodel
            metrics_operation = subclass()  # type: ignore

            if self.model.__class__.get_extension() == metrics_operation.model_type_extension:
                # If its the metamodel that math the model, calculate the metrics
                # Then we calculate the metrics for each metamodel
                sub_metric = subclass()  # type: ignore
                sub_metric.filter = self.filter
                self.result.extend(sub_metric.calculate_metamodel_metrics(model))
            else:
                # If not, search a transformation, transform and call the calutation
                m_to_m = self._search_transformations(self.model.__class__.get_extension(), 
                                                      metrics_operation.model_type_extension)
                dest_model = m_to_m(self.model).transform()
                sub_metric = subclass()  # type: ignore
                sub_metric.filter = self.filter
                self.result.extend(sub_metric.calculate_metamodel_metrics(dest_model))
        return self

    def _search_transformations(self, orig: str, dest: str) -> Type[ModelToModel]:
        try:
            for m_to_m in ModelToModel.__subclasses__():
                _orig = m_to_m.get_source_extension()
                _dest = m_to_m.get_destination_extension()
                if (_orig == orig and _dest == dest):
                    return m_to_m
        except FlamaException:
            LOGGER.exception("No transformation found that is required in the Metrics operation")
        raise FlamaException(f"No transformation found for {orig} -> {dest}")

    def get_result(self) -> list[dict[str, Any]]:
        return self.result

    def get_result_json(self) -> str: 
        return json.dumps(self.get_result(), indent=4) 

    @staticmethod
    def get_ratio(collection1: Collection[Any], 
                  collection2: Collection[Any], 
                  precision: int = 4) -> float:
        if not collection1 and not collection2:
            # Handle case where both collections are empty
            return 0.0
        if not collection2:
            return 0.0
        return float(round(len(collection1) / len(collection2), precision))

    @staticmethod
    def construct_result(name: str,
                         doc: str,
                         result: Any, 
                         size: Optional[int] = None, 
                         ratio: Optional[float] = None,
                         parent: Optional[Any] = None,
                         level: int = 0
                         ) -> dict[str, Any]:  
        # pylint: disable=too-many-arguments
        """Constructs a dictionary with named keys from the provided values.

        Arguments:
            name: The property name.
            doc: The description of the property.
            result: The value(s).
            size (optional): The length of the values (number of values).
            ratio (optional): The percentage of values with regards the total number of possible 
            values.
            parent (optional): The parent metrics of which this metric is based on 
            (for organization purposes).
            level (optional): The number of ancestors of this metrics (for organization purposes).
        """
        return {
            "name": name,
            "documentation": doc,
            "result": result,
            "size": size,
            "ratio": ratio,
            "parent": parent,
            "level": level
        }



from typing import Any
from enum import Enum


class ASTOperation(Enum):
    REQUIRES = 'REQUIRES'
    EXCLUDES = 'EXCLUDES'
    AND = 'AND'
    OR = 'OR'
    XOR = 'XOR'
    IMPLIES = 'IMPLIES'
    NOT = 'NOT'
    EQUIVALENCE = 'EQUIVALENCE'
    # Comparison operators
    EQUALS = 'EQUALS'
    LOWER = 'LOWER'
    GREATER = 'GREATER'
    LOWER_EQUALS = 'LOWER_EQUALS'
    GREATER_EQUALS = 'GREATER_EQUALS'
    NOT_EQUALS = 'NOT_EQUALS'
    # Arithmetic operators
    ADD = 'ADD'
    SUB = 'SUB'
    MUL = 'MUL'
    DIV = 'DIV'
    # Aggregation operators
    SUM = 'SUM'
    AVG = 'AVG'
    # Set operators
    LEN = 'LEN'
    # Numeric agregation operators
    FLOOR = 'FLOOR'
    CEIL = 'CEIL'


class Node:

    def __init__(self, data: Any, left: 'Node' = None, right: 'Node' = None):  # type: ignore
        self.data = data
        self.left = left
        self.right = right

    def is_term(self) -> bool:
        return not self.is_op()

    def is_op(self) -> bool:
        return isinstance(self.data, ASTOperation)

    def is_unary_op(self) -> bool:
        return self.is_op() and self.data in [ASTOperation.NOT]

    def is_unique_term(self) -> bool:
        return not self.is_op() and self.left is None

    def is_binary_op(self) -> bool:
        return not self.is_unique_term() and not self.is_unary_op()

    def __str__(self) -> str:
        data = self.data.value if self.is_op() else safename(str(self.data))
        if self.left and self.right:
            return f'{data}[{self.left}][{self.right}]'
        if self.left and not self.right:
            return f'{data}[{self.left}][]'
        if not self.left and self.right:
            return f'{data}[][{self.right}]'
        return str(data)

    @staticmethod
    def _get_pretty_str_node(node: 'Node') -> str:
        res = ''
        if node.is_op():
            res = f'{node.pretty_str()}'
            if node.is_binary_op():
                res = f'({res})'
        else:
            res = str(node)
        return res

    def pretty_str(self) -> str:
        data = self.data.value if self.is_op() else safename(str(self.data))
        left = Node._get_pretty_str_node(self.left) if self.left is not None else ''
        right = Node._get_pretty_str_node(self.right) if self.right is not None else ''

        if self.is_unique_term():
            res = f'{data}'
        elif self.is_unary_op():
            res = f'{data} {left}'
        else:  # binary operation
            res = f'{left} {data} {right}'
        return res


class AST:
    """Abstract Syntax Tree (AST) to store constraints."""

    def __init__(self, root: Node):
        self.root = root

    @classmethod
    def create_simple_binary_operation(cls, operation: ASTOperation,
                                       left: str, right: str) -> 'AST':
        return cls(Node(operation, Node(left), Node(right)))

    @classmethod
    def create_simple_unary_operation(cls, operation: ASTOperation, elem: str) -> 'AST':
        return cls(Node(operation, Node(elem)))

    @classmethod
    def create_binary_operation(cls, operation: ASTOperation, left: Node, right: Node) -> 'AST':
        return cls(Node(operation, left, right))

    @classmethod
    def create_unary_operation(cls, operation: ASTOperation, elem: Node) -> 'AST':
        return cls(Node(operation, elem))

    def to_cnf(self) -> 'AST':
        return convert_into_cnf(self)

    def get_clauses(self) -> list[list[Any]]:
        ast = self.to_cnf()
        return get_clauses(ast)

    def __str__(self) -> str:
        return str(self.root)

    def pretty_str(self) -> str:
        return self.root.pretty_str()


def convert_into_nnf(ast: AST) -> AST:
    ast = simplify_formula(ast)
    return propagate_negation(ast.root)


def convert_into_cnf(ast: AST) -> AST:
    ast = simplify_formula(ast)
    ast = propagate_negation(ast.root)
    return to_cnf(ast)


def simplify_formula(ast: AST) -> AST:
    """Convert a propositional formula to an equivalent formula without '=>', '<=>', 'XOR',
    'REQUIRES', 'EXCLUDES'.

    Adapted from [Büning, Hans Kleine, and Theodor Lettmann.
    Propositional logic: deduction and algorithms. Vol. 48. Cambridge University Press, 1999.]
    """
    logic_op = ast.root.data
    left = ast.root.left
    right = ast.root.right
    if logic_op in (ASTOperation.REQUIRES, ASTOperation.IMPLIES):
        # Replace P => Q with !P v Q.
        left = simplify_formula(AST(left)).root
        right = simplify_formula(AST(right)).root
        result = AST.create_binary_operation(ASTOperation.OR, Node(ASTOperation.NOT, left), right)
    elif logic_op == ASTOperation.EXCLUDES:
        # Replace P EXCLUDES Q with !P v !Q.
        left = simplify_formula(AST(left)).root
        right = simplify_formula(AST(right)).root
        result = AST.create_binary_operation(ASTOperation.OR, Node(ASTOperation.NOT, left), 
                                             Node(ASTOperation.NOT, right))
    elif logic_op == ASTOperation.EQUIVALENCE:
        # Replace P <=> Q with P => Q ∧ Q => P.
        left = simplify_formula(AST.create_binary_operation(ASTOperation.IMPLIES,
                                                            left, right)).root
        right = simplify_formula(AST.create_binary_operation(ASTOperation.IMPLIES,
                                                             right, left)).root
        result = AST.create_binary_operation(ASTOperation.AND, left, right)
    elif logic_op == ASTOperation.XOR:
        # Replace P XOR Q with (P ∧ !Q) v (!P ∧ Q).
        left = simplify_formula(AST.create_binary_operation(ASTOperation.AND,
                                                            left,
                                                            Node(ASTOperation.NOT, right))).root
        left = simplify_formula(AST.create_binary_operation(ASTOperation.AND,
                                                            Node(ASTOperation.NOT, left),
                                                            right)).root
        result = AST.create_binary_operation(ASTOperation.OR, left, right)
    elif logic_op == ASTOperation.AND:
        left = simplify_formula(AST(left)).root
        right = simplify_formula(AST(right)).root
        result = AST.create_binary_operation(ASTOperation.AND, left, right)
    elif logic_op == ASTOperation.OR:
        left = simplify_formula(AST(left)).root
        right = simplify_formula(AST(right)).root
        result = AST.create_binary_operation(ASTOperation.OR, left, right)
    elif logic_op == ASTOperation.NOT:
        left = simplify_formula(AST(left)).root
        result = AST.create_unary_operation(ASTOperation.NOT, left)
    else:
        result = ast
    return result


def to_cnf(formula: AST) -> AST:
    """Convert a propositional formula in NNF to an equivalent formula in conjunctive normal form.

    Adapted and fixed from [Alexander Knüppel. The Role of Complex Constraints in Feature Modeling.
    Master's Thesis. 2016].
    """
    res = formula
    node = formula.root
    if node.data == ASTOperation.AND:
        res = AST.create_binary_operation(ASTOperation.AND,
                                          to_cnf(AST(node.left)).root,
                                          to_cnf(AST(node.right)).root)
    elif node.data == ASTOperation.OR:
        node.left = to_cnf(AST(node.left)).root
        node.right = to_cnf(AST(node.right)).root
        if node.left.data == ASTOperation.AND:
            res = AST.create_binary_operation(ASTOperation.AND,
                                              AST.create_binary_operation(ASTOperation.OR,
                                                                          node.left.left,
                                                                          node.right).root,
                                              AST.create_binary_operation(ASTOperation.OR,
                                                                          node.left.right,
                                                                          node.right).root)
            res = to_cnf(res)
        elif node.right.data == ASTOperation.AND:
            res = AST.create_binary_operation(ASTOperation.AND,
                                              AST.create_binary_operation(ASTOperation.OR,
                                                                          node.left,
                                                                          node.right.left).root,
                                              AST.create_binary_operation(ASTOperation.OR,
                                                                          node.left,
                                                                          node.right.right).root)
            res = to_cnf(res)
        else:
            res = AST.create_binary_operation(ASTOperation.OR, node.left, node.right)
    return res


def to_nnf(ast: AST) -> AST:
    """Convert a simplified propositional formula to an equivalent formula in negation normal form.

    A simplified formula only contains '∧', 'v', and 'not'.

    Adapted from [Alexander Knüppel. The Role of Complex Constraints in Feature Modeling.
    Master's Thesis. 2016].
    """
    return propagate_negation(ast.root)


def propagate_negation(node: Node, negated: bool = False) -> AST:
    result = None
    if node.data == ASTOperation.NOT:
        negated = not negated
        result = propagate_negation(node.left, negated)
    elif node.data == ASTOperation.AND:
        if negated:
            result = AST.create_binary_operation(ASTOperation.OR,
                                                 propagate_negation(node.left, negated).root,
                                                 propagate_negation(node.right, negated).root)
        else:
            result = AST.create_binary_operation(ASTOperation.AND,
                                                 propagate_negation(node.left, negated).root,
                                                 propagate_negation(node.right, negated).root)
    elif node.data == ASTOperation.OR:
        if negated:
            result = AST.create_binary_operation(ASTOperation.AND,
                                                 propagate_negation(node.left, negated).root,
                                                 propagate_negation(node.right, negated).root)
        else:
            result = AST.create_binary_operation(ASTOperation.OR,
                                                 propagate_negation(node.left, negated).root,
                                                 propagate_negation(node.right, negated).root)
    else:
        if negated:
            result = AST.create_unary_operation(ASTOperation.NOT, node)
        else:
            result = AST(node)
    return result


def get_clauses(ast: AST) -> list[list[Any]]:
    """Return the list of clauses represented by the AST root node in conjunctive normal form."""
    node = ast.root
    result = []
    if node.is_term():
        result = [[node.data]]
    elif node.data == ASTOperation.NOT:
        result = [['-' + node.left.data]]
    elif node.data == ASTOperation.OR:
        result = [get_clause_from_or_node(node)]
    elif node.data == ASTOperation.AND:  # Each AND gives us two clauses
        result.extend(get_clauses(AST(node.left)))
        result.extend(get_clauses(AST(node.right)))
    return result


def get_clause_from_or_node(node: Node) -> list[Any]:
    clause = []
    if node.left.is_op() and node.left.data == ASTOperation.OR:
        # recursive OR belongs to the same clause
        clause.extend(get_clause_from_or_node(node.left))
    else:
        term = node.left.data if node.left.is_term() else f'-{node.left.left.data}'
        clause.append(term)
    if node.right.is_op() and node.right.data == ASTOperation.OR:
        # recursive OR belongs to the same clause
        clause.extend(get_clause_from_or_node(node.right))
    else:
        term = node.right.data if node.right.is_term() else f'-{node.right.left.data}'
        clause.append(term)
    return clause


def safename(name: str) -> str:
    return f'"{name}"' if ' ' in name else name



from abc import abstractmethod

from flamapy.core.operations import Operation


class Variability(Operation):

    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_variability(self) -> float:
        pass



class FlamaException(Exception):
    pass


class ParsingException(Exception):
    pass


class PluginNotFound(FlamaException):
    pass


class OperationNotFound(FlamaException):
    pass


class TransformationNotFound(FlamaException):
    pass


class ElementNotFound(FlamaException):
    pass


class DuplicatedFeature(FlamaException):
    pass


class ConfigurationNotFound(FlamaException):
    pass



import csv

from flamapy.core.transformations.text_to_model import TextToModel

from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration
from flamapy.core.utils import file_exists
from flamapy.core.exceptions import ConfigurationNotFound


class ConfigurationBasicReader(TextToModel):
    @staticmethod
    def get_source_extension() -> str:
        return 'csvconf'

    def __init__(self, path: str) -> None:
        self._path = path

    def transform(self) -> Configuration:
        csv_reader = self.get_configuration_from_csv(self._path)
        elements = {}
        for row in csv_reader:
            # Assuming that row[1] is supposed to represent a boolean value
            # Convert 'true'/'false' strings to actual boolean values
            elements[row[0]] = row[1].lower() == 'true'
        return Configuration(elements)

    def get_configuration_from_csv(self, path: str) -> list[list[str]]:
        # Returns a list of list
        if not file_exists(path):
            raise ConfigurationNotFound

        with open(path, 'r', encoding='utf-8') as csvfile:
            csvreader = list(csv.reader(csvfile))

        return csvreader



from abc import abstractmethod
from typing import Any

from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration
from flamapy.core.operations import Operation


class Filter(Operation):

    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def set_configuration(self, configuration: Configuration) -> None:
        pass

    @abstractmethod
    def get_filter_products(self) -> list[Any]:
        pass



import inspect
import logging
from importlib import import_module
from pkgutil import iter_modules
from types import ModuleType
from typing import Any, Optional, Protocol, Type, runtime_checkable, cast

from flamapy.core.config import PLUGIN_PATHS
from flamapy.core.exceptions import OperationNotFound
from flamapy.core.exceptions import TransformationNotFound
from flamapy.core.exceptions import ConfigurationNotFound
from flamapy.core.models import VariabilityModel
from flamapy.core.operations import Operation
from flamapy.core.plugins import (
    Operations,
    Plugin,
    Plugins
)
from flamapy.core.transformations import Transformation
from flamapy.core.transformations.text_to_model import TextToModel
from flamapy.core.transformations.model_to_model import ModelToModel
from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration


LOGGER = logging.getLogger('discover')


@runtime_checkable
class OperationWithConfiguration(Protocol):
    def set_configuration(self, configuration: Configuration) -> None:
        pass


def filter_modules_from_plugin_paths() -> list[ModuleType]:
    results: list[ModuleType] = []
    for path in PLUGIN_PATHS:
        try:
            module: ModuleType = import_module(path)
            results.append(module)
        except ModuleNotFoundError:
            LOGGER.exception('ModuleNotFoundError %s', path)
    return results


class DiscoverMetamodels:
    def __init__(self) -> None:
        self.module_paths = filter_modules_from_plugin_paths()
        self.plugins: Plugins = self.discover()

    def search_classes(self, module: ModuleType) -> list[Any]:
        classes = []
        for _, file_name, ispkg in iter_modules(
            module.__path__, module.__name__ + '.'
        ):
            if ispkg:
                classes += self.search_classes(import_module(file_name))
            else:
                _file = import_module(file_name)
                classes += inspect.getmembers(_file, inspect.isclass)
        return classes

    def discover(self) -> Plugins:
        plugins = Plugins()
        for pkg in self.module_paths:
            for _, plugin_name, ispkg in iter_modules(
                pkg.__path__, pkg.__name__ + '.'
            ):
                if not ispkg:
                    continue
                module = import_module(plugin_name)
                plugin = Plugin(module=module)

                classes = self.search_classes(module)

                for _, _class in classes:
                    if not _class.__module__.startswith(module.__package__):
                        continue  # Exclude modules not in current package
                    inherit = _class.mro()

                    if Operation in inherit:
                        plugin.append_operation(_class)
                    elif Transformation in inherit:
                        plugin.append_transformations(_class)
                    elif VariabilityModel in inherit:
                        plugin.variability_model = _class
                plugins.append(plugin)
        return plugins

    def reload(self) -> None:
        self.plugins = self.discover()

    def get_operations(self) -> list[Type[Operation]]:
        """ Get the operations for all modules """
        operations: list[Type[Operation]] = []
        for plugin in self.plugins:
            operations += plugin.operations
        return operations

    def get_name_operations(self) -> list[str]:
        operations = []
        for operation in self.get_operations():
            operations.append(operation.__name__)
            base = operation.__base__.__name__ if operation.__base__ else ''
            if base != 'ABC':
                operations.append(base)

        return operations

    def get_transformations(self) -> list[Type[Transformation]]:
        """ Get transformations for all modules """
        transformations: list[Type[Transformation]] = []
        for plugin in self.plugins:
            transformations += plugin.transformations
        return transformations

    def get_transformations_t2m(self) -> list[Type[TextToModel]]:
        """ Get t2m transformations for all modules """

        transformations: list[Type[TextToModel]] = []
        for plugin in self.plugins:
            transformations += [
                t for t in plugin.transformations if issubclass(t, TextToModel)
            ]
        return transformations

    def get_transformations_m2m(self) -> list[Type[ModelToModel]]:
        """ Get m2m transformations for all modules """

        transformations: list[Type[ModelToModel]] = []
        for plugin in self.plugins:
            transformations += [
                t for t in plugin.transformations if issubclass(t, ModelToModel)
            ]
        return transformations

    def get_operations_by_plugin(self, plugin_name: str) -> Operations:
        return self.plugins.get_operations_by_plugin_name(plugin_name)

    def get_plugins_with_operation(self, operation_name: str) -> list[Plugin]:
        return [
            plugin for plugin in self.plugins
            if operation_name in self.get_name_operations_by_plugin(plugin.name)
        ]

    def get_name_operations_by_plugin(self, plugin_name: str) -> list[str]:
        operations = []
        for operation in self.get_operations_by_plugin(plugin_name):
            operations.append(operation.__name__)
            base = operation.__base__.__name__ if operation.__base__ else ''
            if base != 'ABC':
                operations.append(base)

        return operations

    def get_variability_models(self) -> list[VariabilityModel]:
        return self.plugins.get_variability_models()

    def get_plugins(self) -> list[str]:
        return self.plugins.get_plugin_names()

    def use_transformation_m2t(self, src: VariabilityModel, dst: str) -> str:
        plugin = self.plugins.get_plugin_by_variability_model(src)
        return plugin.use_transformation_m2t(src, dst)

    def use_transformation_t2m(self, src: str, dst: str) -> VariabilityModel:
        plugin = self.plugins.get_plugin_by_extension(dst)
        return plugin.use_transformation_t2m(src)

    def use_transformation_m2m(self, src: VariabilityModel, dst: str) -> VariabilityModel:
        plugin = self.plugins.get_plugin_by_extension(dst)
        return plugin.use_transformation_m2m(src, dst)

    def get_operation(self, src: VariabilityModel, operation_name: str) -> Operation:
        plugin = self.plugins.get_plugin_by_variability_model(src)
        return plugin.get_operation(operation_name)

    def use_operation(self, src: VariabilityModel, operation_name: str) -> Operation:
        plugin = self.plugins.get_plugin_by_variability_model(src)
        operation = plugin.get_operation(operation_name)
        return plugin.use_operation(operation, src)

    # pylint: disable=too-many-arguments
    def use_operation_from_vm(
        self,
        operation_name: str,
        vm_orig: VariabilityModel,
        plugin_name: Optional[str] = None,
        configuration_file: Optional[str] = None,
        is_full: Optional[bool] = False
    ) -> Any:

        if operation_name not in self.get_name_operations():
            raise OperationNotFound()
        vm_temp = vm_orig
        if plugin_name is not None:
            plugin = self.plugins.get_plugin_by_name(plugin_name)
            #vm_temp = plugin.use_transformation_t2m(file)
        else:
            #vm_temp = self.__transform_to_model_from_file(file)
            plugin = self.plugins.get_plugin_by_extension(
                vm_orig.get_extension())

            if operation_name not in self.get_name_operations_by_plugin(plugin.name):
                transformation_way = self.__search_transformation_way(
                    plugin, operation_name)

                for (_, dst) in transformation_way:
                    _plugin = self.plugins.get_plugin_by_extension(dst)
                    vm_temp = _plugin.use_transformation_m2m(vm_temp, dst)
                    plugin = _plugin

        operation = plugin.get_operation(operation_name)
        if isinstance(operation, OperationWithConfiguration):
            if configuration_file is None:
                raise ConfigurationNotFound()
            configuration = self.__transform_to_model_from_file(configuration_file)
            configuration = cast(Configuration, configuration)
            is_full_value = is_full if is_full is not None else False
            configuration.set_full(is_full_value)
            operation.set_configuration(configuration)

        operation = plugin.use_operation(operation, vm_temp)

        return operation.get_result()

    # pylint: disable=too-many-arguments
    def use_operation_from_file(
        self,
        operation_name: str,
        file: str,
        plugin_name: Optional[str] = None,
        configuration_file: Optional[str] = None,
        is_full: Optional[bool] = False
    ) -> Any:

        if operation_name not in self.get_name_operations():
            raise OperationNotFound()

        if plugin_name is not None:
            plugin = self.plugins.get_plugin_by_name(plugin_name)
            vm_temp = plugin.use_transformation_t2m(file)
        else:
            vm_temp = self.__transform_to_model_from_file(file)
            plugin = self.plugins.get_plugin_by_extension(
                vm_temp.get_extension())

            if operation_name not in self.get_name_operations_by_plugin(plugin.name):
                transformation_way = self.__search_transformation_way(
                    plugin, operation_name)

                for (_, dst) in transformation_way:
                    _plugin = self.plugins.get_plugin_by_extension(dst)
                    vm_temp = _plugin.use_transformation_m2m(vm_temp, dst)
                    plugin = _plugin

        operation = plugin.get_operation(operation_name)
        if isinstance(operation, OperationWithConfiguration):
            if configuration_file is None:
                raise ConfigurationNotFound()
            configuration = self.__transform_to_model_from_file(configuration_file)
            configuration = cast(Configuration, configuration)
            is_full_value = is_full if is_full is not None else False
            configuration.set_full(is_full_value)
            operation.set_configuration(configuration)

        operation = plugin.use_operation(operation, vm_temp)
        return operation.get_result()

    def __transform_to_model_from_file(self, file: str) -> VariabilityModel:
        t2m_transformations = self.get_transformations_t2m()
        extension = file.split('.')[-1]
        t2m_filters = filter(
            lambda t2m: t2m.get_source_extension() == extension,
            t2m_transformations
        )

        t2m = next(t2m_filters, None)
        if t2m is None:
            raise TransformationNotFound()

        return t2m(file).transform()

    def __search_transformation_way(
        self,
        plugin: Plugin,
        operation_name: str
    ) -> list[tuple[str, str]]:
        '''
        Search way to reach plugin with operation_name using m2m transformations
        '''
        way: list[tuple[str, str]] = []

        plugins_with_operation = self.get_plugins_with_operation(
            operation_name)
        m2m_transformations = self.get_transformations_m2m()

        input_extension = plugin.get_extension()

        def __search_recursive_way(
            input_extension: str,
            output_extension: str,
            tmp_way: list[tuple[str, str]]
        ) -> list[tuple[str, str]]:

            for m2m in m2m_transformations:
                in_m2m = m2m.get_source_extension()
                out_m2m = m2m.get_destination_extension()

                if out_m2m == output_extension:
                    _next = (in_m2m, out_m2m)

                    if _next in tmp_way:
                        continue

                    tmp_way.insert(0, _next)

                    if input_extension == in_m2m:
                        return tmp_way

                    return __search_recursive_way(input_extension, in_m2m, tmp_way)

            return tmp_way

        for _plugin in plugins_with_operation:
            output_extension = _plugin.get_extension()
            way = __search_recursive_way(input_extension, output_extension, [])
            if way and output_extension == way[-1][1]:
                return way

        raise NotImplementedError('Way to execute operation not found')



from abc import abstractmethod

from flamapy.core.operations import Operation
from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration


class Configurations(Operation):

    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_configurations(self) -> list[Configuration]:
        pass



from abc import ABC, abstractmethod


class VariabilityModel(ABC):

    @staticmethod
    @abstractmethod
    def get_extension() -> str:
        """ Plugin file extension """


class VariabilityElement():

    def __init__(self, name: str) -> None:
        self.name = name



from types import ModuleType
from typing import Any, Callable, Type, cast
from collections import UserList

from flamapy.core.exceptions import (
    OperationNotFound,
    PluginNotFound,
    TransformationNotFound,
)
from flamapy.core.models import VariabilityModel
from flamapy.core.operations import Operation
from flamapy.core.transformations import (
    TextToModel,
    Transformation,
    ModelToText,
    ModelToModel,
)
from flamapy.core.utils import extract_filename_extension


class Transformations(UserList[Type[Transformation]]):
    data: list[Type[Transformation]]


class Operations(UserList[Type[Operation]]):
    data: list[Type[Operation]]

    def search_by_name(self, name: str) -> Type[Operation]:
        candidates: filter[Type[Operation]] = filter(
            lambda op: name in [op.__name__, op.__base__.__name__ if op.__base__ else ''],
            self.data
        )

        try:
            operation = next(candidates, None)
        except StopIteration:
            raise OperationNotFound from StopIteration

        if not operation:
            raise OperationNotFound
        return operation


class Plugin:

    def __init__(self, module: ModuleType) -> None:
        self.module: ModuleType = module
        self.variability_model: VariabilityModel = None  # type: ignore
        self.operations: Operations = Operations()
        self.transformations: Transformations = Transformations()

    def __get_transformation(
        self,
        filter_transformation: Callable[..., bool]
    ) -> Type[Transformation]:
        candidates = filter(filter_transformation, self.transformations)
        try:
            transformation = next(candidates, None)
        except StopIteration:
            raise TransformationNotFound from StopIteration

        if not transformation:
            raise TransformationNotFound
        return transformation

    def append_operation(self, operation: Type[Operation]) -> None:
        self.operations.append(operation)

    def append_transformations(self, transformation: Type[Transformation]) -> None:
        self.transformations.append(transformation)

    def get_operation(self, name: str) -> Operation:
        operation = self.operations.search_by_name(name)
        return operation()

    def use_operation(self, operation: Operation, src: VariabilityModel) -> Operation:
        return operation.execute(model=src)

    def use_transformation_t2m(self, src: str) -> VariabilityModel:
        extension = extract_filename_extension(src)

        def filter_transformations(transformation: Type[Transformation]) -> bool:
            return issubclass(transformation, TextToModel) and\
                transformation.get_source_extension() == extension

        transformation: Type[TextToModel] = cast(
            Type[TextToModel],
            self.__get_transformation(filter_transformations)
        )
        result = transformation(src)
        return result.transform()

    def use_transformation_m2t(self, src: VariabilityModel, dst: str) -> str:
        extension = extract_filename_extension(dst)

        def filter_transformations(transformation: Type[Transformation]) -> bool:
            return issubclass(transformation, ModelToText) and\
                transformation.get_destination_extension() == extension

        transformation: Type[ModelToText] = cast(
            Type[ModelToText],
            self.__get_transformation(filter_transformations)
        )
        result = transformation(path=dst, source_model=src)
        return result.transform()

    def use_transformation_m2m(
        self,
        src: VariabilityModel,
        dst: str
    ) -> VariabilityModel:
        def filter_transformations(transformation: Type[Transformation]) -> bool:
            return issubclass(transformation, ModelToModel) and\
                transformation.get_destination_extension() == dst and\
                transformation.get_source_extension() == src.get_extension()

        transformation: Type[ModelToModel] = cast(
            Type[ModelToModel],
            self.__get_transformation(filter_transformations)
        )
        result = transformation(src)
        return result.transform()

    def get_extension(self) -> str:
        return self.variability_model.get_extension()

    @property
    def name(self) -> str:
        return self.module.__name__.split('.')[-1]

    def get_stats(self) -> dict[str, Any]:
        return {
            'amount_operations': len(self.operations),
            'amount_transformations': len(self.transformations),
            'variability_model': bool(self.variability_model)
        }


class Plugins(UserList[Plugin]):
    data: list[Plugin]

    def __get_plugin_by_filter(self, plugin_filter: Callable[[Plugin], bool]) -> Plugin:
        candidates = filter(plugin_filter, self.data)
        try:
            plugin = next(candidates)
        except StopIteration:
            raise PluginNotFound from StopIteration
        return plugin

    def get_plugin_by_name(self, name: str) -> Plugin:

        def plugin_filter(plugin: Plugin) -> bool:
            return plugin.name == name

        return self.__get_plugin_by_filter(plugin_filter)

    def get_plugin_by_variability_model(
        self,
        variability_model: VariabilityModel
    ) -> Plugin:

        def plugin_filter(plugin: Plugin) -> bool:

            return isinstance(variability_model,
                              plugin.variability_model)  # type: ignore
            # TODO: Check if this error can be ignored

        return self.__get_plugin_by_filter(plugin_filter)

    def get_plugin_by_extension(self, extension: str) -> Plugin:

        def plugin_filter(plugin: Plugin) -> bool:
            return extension == plugin.get_extension()

        return self.__get_plugin_by_filter(plugin_filter)

    def get_plugin_names(self) -> list[str]:
        return [plugin.name for plugin in self.data]

    def get_variability_models(self) -> list[VariabilityModel]:
        return [plugin.variability_model for plugin in self.data]

    def get_operations_by_plugin_name(self, plugin_name: str) -> Operations:
        try:
            plugin = self.get_plugin_by_name(plugin_name)
            return plugin.operations
        except PluginNotFound:
            return Operations()

    def get_stats(self) -> dict[str, Any]:
        stats: dict[str, Any] = {'amount_plugins': len(self.data)}
        for plugin in self.data:
            stats[plugin.name] = plugin.get_stats()
        return stats
